package com.cabintechglobal.sampleratefix;

/**
 * SampleRateFix
 * Author: Mark McMillan, CabinTech Global LLC
 * 
 * This code was developed out of a desperate need to repair audio recordings made on an HD24 digital
 * recorder. The recorder was fed 3 ADAT inputs (24 channels) from a Yamaha M7CL digital mixer. The
 * HD24 was setup to sync on its INTERNAL clock at 48kHz. The Yamaha was generating ADAT clocking at
 * 44.1kHz. The HD24 should have been setup to sync on the optical (ADAT) input stream. Operator error
 * (and poor design of the HD24 which should not have allowed this configuration). See the discussion
 * in the HD24 users group at https://groups.yahoo.com/neo/groups/hd24/conversations/topics/34589.
 * 
 * The result was 24 tracks (80 minutes each) of recordings which were sampled about 9% too fast. That
 * means that every so often, the HD24 took a sample before the inputs had changed to the next value.
 * That resulted in duplicate values in the recorded data about every 12.3 samples (based on the ratio
 * of 48kHz to 44.1kHz). This resulted in recordings with a distinct high-frequency distortion 
 * perceived as a "ringing" around the edges of the sound, destroying the quality of the recordings.
 * 
 * This program was written to seek out and destroy the erroneous duplicate samples, and create a new
 * WAV file (at the original 44.1kHz sampling frequency). The repaired WAV file is exactly what should
 * have been recorded had the Yamaha and HD24 been properly synchronized.
 * 
 * This code is donated to the public domain and may be used for any purpose. Note the use of the
 * WavFile utility code which may have its own license restrictions (see http://www.labbookpages.co.uk/audio/javaWavFiles.html).
 * The WavFile utility was slightly modified for this application (see WavFile.java).
 * 
 * Sample usage and output:
 * 
		>java -jar sampleratefix.jar c:\temp\audio\distorted_piano_int24.wav
		 
		Reading file 'c:\temp\audio\distorted_piano_int24.wav'
		Channels: 1
		Sample rate: 48000
		Bits per sample: 24
		Sync starting at sample 10
		Sync confirmed at sample 22
		Processing complete.
		Read 11261348 total samples.
		Found 914813 suspect duplicate samples.
		Found 71 suspect corrupted samples.
		
		Wrote 10346468 output samples.
		Processing time: 0.525 seconds.
		
		Output file (24 bits/sample, mono, 44.1kHz): c:\temp\audio\distorted_piano_int24_repaired.wav
 * 
 */

import java.io.File;
import java.io.IOException;

import com.wavefile.WavFile;
import com.wavefile.WavFileException;

public class SampleRateFix {
	
	private static int MIN_INTERVAL_SIZE = 11;   // Min interval (samples) between bad values
	private static int SYNC_WINDOW_SIZE = 2;     // Size of search window (e.g. variation in location of bad samples)
	private static int[] outBuff = new int[100]; // Buffer of 100 output frames (=100 samples)
	private static int outBuffPtr = 0;           // Next unused position in the outBuff
	private static WavFile wavFileOut = null;
	private static WavFile wavFileIn = null;
	private static int inputNumChannels = 0;
	private static long inputSampleRate = 0;
	private static long inputNumFrames = 0;
	private static long outputNumFrames = 0;
	private static int sampleNum = 0;            // Current sample number (1 based)
	private static double currVal = 0.0;         // Current value read from the input stream
	private static double nextVal = 0.0;         // Next value to become the currVal
	
	private static boolean debug = false;        // Output some debug messages

	
	public static void main(String[] args)
	{
		try
		{
			// Check input args
			if (args.length == 0) {
				System.out.println("usage: sampleratefix <inputfile.wav> [-debug]");
				System.out.println("Repaired file will be written to <inputfile_repaired.wav>");
				System.out.println("");
				System.out.println("Author: Mark McMillan");
				return;
			}
			File inFile = new File(args[0]);
			if (!inFile.exists()) {
				System.out.println("Input file '"+inFile.getAbsolutePath()+"' not found.");
				return;
			}
			
			for (int i=1; i<args.length; i++) {
				if (args[i].equals("-debug")) {
					debug = true;
				}
			}
			
			// Create input/output File objects
			String fName = inFile.getName();
			int idx = fName.lastIndexOf('.');
			if (idx > 0) {
				fName = fName.substring(0,idx);
			}
			fName = fName + "_repaired.wav";
			File outFile = new File(inFile.getParent(), fName);
			
			// Open the input and output WAV files
			System.out.println("Reading file '"+inFile.getAbsolutePath()+"'");
			wavFileIn = WavFile.openWavFile(inFile);

			// Get information about the wav file
			inputNumChannels = wavFileIn.getNumChannels();
			inputSampleRate = wavFileIn.getSampleRate();
			inputNumFrames = wavFileIn.getNumFrames(); // Same as number of samples for 1-channel wav file

			// Display information about the wav file
			System.out.println("Channels: "+inputNumChannels);
			System.out.println("Sample rate: "+inputSampleRate);
			System.out.println("Bits per sample: "+wavFileIn.getValidBits());
			System.out.println("Frames: "+wavFileIn.getNumFrames());
			
			if (inputSampleRate != 48000) { // Sanity check
				System.out.println("ERROR: Input file sample rate must be 48,000.");
				wavFileIn.close();
				return;
			}
			if (inputNumChannels > 1) {
				System.out.println("WARNING: Only channel 1 will be processed. Output file will be mono.");
			}
			
			// Create the output wav file 24 bit mono 44.1kHz. Unfortunately, the WavFile utility requires us to know
			// the number of frames even before we start writing the data. Unless we make 2 passes through the
			// input, we don't know how many frame we will drop, so we don't know yet how much we will write.
			// Best we can do is estimate the number of frames based on the duration of the input wav. So the
			// exact number of samples written may not match the wav file header ("chunk") data. Different tools
			// will handle that problem differently, most will just read all the samples that exist in the file.
			//
			//TODO: Modify WavFile to write zero data-chunck sizes on the first pass, then go back and
			// use random-writes into the file to update those after all the data is written.
			
			double duration = inputNumFrames/inputSampleRate; // Input wav file playing time, in seconds
			double outputFramesEst = duration * 44100;        // Estimate number of output frames
			
			wavFileOut = WavFile.newWavFile(outFile, 1, (long)outputFramesEst, 24, 44100);
			
			// Ready, set, ...
			int dupsFound = 0;
			int junkFound = 0;
			long startTime = System.currentTimeMillis();
			
			//skipSamples(430);
			
			// Prime the sample queue
			getNextSample();
			getNextSample();
			sampleNum = 1;
			
			boolean inSync = false;
			while (!inSync) {
			
				// Scan for next duplicate value
				while (true) {
					putSample(currVal);   // Write last sample to output stream
					
					if (currVal == nextVal) {
						break;
					}
					if (sampleNum > inputSampleRate) {
						// Should find a dup within 1 second worth of samples
						System.out.println("Did not find sync within first 1 second, cannot fix.");
						return;
					}
					getNextSample(); // Shift and read next sample
				}
				
				// Verify we have at least 2 cycles of sync
				
				skipSamples(MIN_INTERVAL_SIZE);
				if (isSyncPoint()) {
					getNextSample(); // Skip the dup
					skipSamples(MIN_INTERVAL_SIZE);
					if (isSyncPoint()) {
						// OK, we have a valid sync going
						getNextSample(); // Skip the dup
						inSync = true;
					}
				}
				
				// Repeat single-sample scan for next duplicate
			}
			
			// At this point we have sync going.
			System.out.println("Sync starting at sample "+sampleNum);
			
			dupsFound = 2; // 2 sync cycles already processed
			
			// Look for dups every 12 or 13 samples
			try {
				while (true) { // Loop until end of data (exception from read code).
					skipSamples(MIN_INTERVAL_SIZE);
					if (isSyncPoint()) {
						getNextSample(); // Skip the dup
						dupsFound++;
						continue;
					}
					
					// Found no dup in the window. Most likely the bad sample is corrupted instead of
					// duplicated (sampled at time of signal transitions). Best algorithm would be to examine
					// the samples in the window and throw out the one that is most distant from the others.
					// That would require backing up and we already wrote some of the window samples to the
					// output. Instead we arbitrarily decide that the last sample is the bad one and we
					// to not write it to the output stream.
					
					junkFound++;
					getNextSample(); // Skip the junk
					if (debug) System.out.println("Corruption suspected at sample "+sampleNum+" ("+getSeconds(sampleNum,inputSampleRate)+" seconds)");
				}
			}
			catch (WavFileException ex) {
				flushOutput();
				wavFileOut.close();
				wavFileIn.close();
				System.out.println("Processing complete.");
				System.out.println("Read "+sampleNum+" total samples ("+getSeconds(sampleNum,inputSampleRate)+" seconds @"+inputSampleRate+" Hz.)");
				System.out.println("Found "+dupsFound+" suspect duplicate samples.");
				System.out.println("Found "+junkFound+" suspect corrupted samples.");
				System.out.println();
				System.out.println("Wrote "+outputNumFrames+" output samples ("+getSeconds((int)outputNumFrames,44100)+" seconds @44100 Hz.)");
				System.out.println("Output file (24 bits/sample, mono, 44.1kHz): "+outFile.getAbsolutePath());
				System.out.println("");
				System.out.println("Processing time: "+(System.currentTimeMillis()-startTime)/1000.0+" seconds.");
				return;
			}
			
		}
		catch (Exception e)
		{
			System.out.println("Sorry, we crashed and burned:");
			System.err.println(e);
		}
	}
	
	/**
	 * Makes the currVal the nextVal, and loads nextVal from the next sample
	 * in the input stream (channel 1 only, other channels are skipped).
	 * @return
	 * @throws IOException
	 * @throws WavFileException
	 */
	private static void getNextSample() throws IOException, WavFileException {
		currVal = nextVal;
		nextVal = wavFileIn.readSample(); // Read ahead one sample
		for (int chan=1;chan<inputNumChannels;chan++) { // discard other channels
			wavFileIn.readSample();
		}
		sampleNum++;
	}
	
	/**
	 * Skips (advances the input wav reader) by the number of samples given. Each
	 * skipped sample is written to the output stream unaltered.
	 * @param cnt
	 * @throws IOException
	 * @throws WavFileException
	 */
	private static void skipSamples(int cnt) throws IOException, WavFileException {
		for (int i=0;i<cnt;i++) {
			putSample(currVal);
			getNextSample();
		}
	}
	
	/**
	 * Returns the time in the wav file of the given sample number, assuming the given sample rate.
	 * @param sample
	 * @param sampleRate
	 * @return
	 */
	private static double getSeconds(int sample, long sampleRate) {
		return sample / sampleRate;
	}
	
	/**
	 * Puts a single sample into the output wav file stream (which is buffered, so this does
	 * not necessarily cause a physical write).
	 * @param outWav
	 * @param sample
	 * @throws WavFileException 
	 * @throws IOException 
	 */
	private static void putSample(double sample) throws IOException, WavFileException {
		//System.out.println("Putting: "+sample);
		outBuff[outBuffPtr] = (int)sample;
		outBuffPtr++;
		if (outBuffPtr >= outBuff.length) {
			// Send this to the file writer. Note that frames==samples for a mono (1-channel) wav file.
			wavFileOut.writeFrames(outBuff, outBuff.length);
			outputNumFrames = outputNumFrames + outBuff.length;
			// Reset buffer for next write
			outBuffPtr = 0;
		}
	}
	
	private static void flushOutput() throws IOException, WavFileException {
		// Write any frames (samples) left in the buffer
		if (outBuffPtr > 0) {
			wavFileOut.writeFrames(outBuff, outBuffPtr);
			outputNumFrames = outputNumFrames + outBuffPtr;
		}
	}
	
	private static boolean isSyncPoint() throws IOException, WavFileException {
		if (currVal == nextVal) return true;
		putSample(currVal);
		getNextSample();
		if (currVal == nextVal) return true;
		return false;
	}

	
}
